//
// This file is part of MorphoGraphX - http://www.MorphoGraphX.org
// Copyright (C) 2012-2015 Richard S. Smith and collaborators.
//
// If you use MorphoGraphX in your work, please cite:
//   http://dx.doi.org/10.7554/eLife.05864
//
// MorphoGraphX is free software, and is licensed under under the terms of the 
// GNU General (GPL) Public License version 2.0, http://www.gnu.org/licenses.
// 
#ifndef SYSTEMPROCESSLOAD_HPP
#define SYSTEMPROCESSLOAD_HPP

#include <Process.hpp>

#include <QObject>

/**
 * \file SystemProcessLoad.hpp
 * This file contains processes defined in MorphoGraphX to load data
 */

class QDialog;
class Ui_LoadStackDialog;
class Ui_LoadMeshDialog;
class Ui_ImportMeshDialog;
class QFile;
class QIODevice;

namespace mgx { namespace process {
  /**
   * This function creates a graph from a triangle soup.
   *
   * \note The graph will not be cleared, the triangles will be added to the
   * graph.
   *
   * \param S Graph that will be updated.
   * \param vertices List of vertices to insert in S
   * \param triangles List of triplet of indices in \c vertices defining
   * triangles
   *
   * \returns true on success, false on failure
   *
   * \ingroup ProcessUtils
   */
  mgx_EXPORT bool meshFromTriangles(vvgraph& S, std::vector<vertex>& vertices,
                                    std::vector<Point3i>& triangles);
  
  mgx_EXPORT QList<int> extractVersion(QIODevice& file);
  
  /**
   * \class StackSwapBytes SystemProcessLoad.hpp <SystemProcessLoad.hpp>
   *
   * Process that swap bytes
   *
   * \ingroup StackProcess
   */
  class mgx_EXPORT StackSwapBytes : public StackProcess {
  public:
    StackSwapBytes(const StackProcess& process) : Process(process), StackProcess(process) {}
  
    bool operator()(const QStringList&)
    {
      if(!checkState().store(STORE_NON_EMPTY))
        return false;
      Store* input = currentStack()->currentStore();
      Store* output = currentStack()->work();
      bool res = (*this)(input, output);
      if(res) {
        input->hide();
        output->show();
      }
      return res;
    }
  
    bool operator()(const Store* input, Store* output);
  
    QString name() const { return "Swap Bytes"; }
    QString description() const { return "Swap the bytes of the values in the stack."; }
    QString folder() const { return "Filters"; }
    QStringList parmNames() const { return QStringList(); }
    QStringList parmDescs() const { return QStringList(); }
    QIcon icon() const { return QIcon(":/images/SwapBytes.png"); }
  };
  
  /**
   * \class StackImport SystemProcessLoad.hpp <SystemProcessLoad.hpp>
   *
   * Import images forming a stack.
   *
   * \ingroup StackProcess
   */
  class mgx_EXPORT StackImport : public QObject, public StackProcess {
    Q_OBJECT
  public:
    StackImport(const StackProcess& proc) : Process(proc), QObject(), StackProcess(proc) {}
  
    bool initialize(QStringList& parms, QWidget* parent);
  
    bool operator()(const QStringList& parms)
    {
      int stackId = parms[0].toInt();
      QString storeName = parms[1];
      Point3f step(parms[2].toFloat(), parms[3].toFloat(), parms[4].toFloat());
      float brightness(parms[5].toFloat());
      QString filename(parms[6]);
  
      // Get stack and store
      Stack* stk = currentStack();
      if(stackId != -1)
        stk = stack(parms[0].toInt());
      if(!checkState().stack(STACK_ANY, stackId))
        return false;
  
      Store* store = stk->currentStore();
      if(storeName == "Main")
        store = stk->main();
      else if(storeName == "Work")
        store = stk->work();
  
      return (*this)(stk, store, step, brightness, filename);
    }
    /**
     * Import the stack
     * \param stack Stack that will contain the image
     * \param store Store that will contain the image
     * \param filenames Either a single text file containing all the
     * parameters, or a list of image files, one per z slice
     * \param brightness Multiplication coefficient to convert less than 16
     * bits images to 16 bits
     * \param step Size of a voxel
     */
    bool operator()(Stack* stack, Store* store, Point3f step, float brightness, QString filename);
  
    QString folder() const { return "System"; }
    QString name() const { return "Import"; }
    QString description() const { return "Import stack from a series of images"; }
    QStringList parmNames() const { return QStringList() 
      << "Stack" << "Store" << "X Step" << "Y Step" << "Z Step" << "Brightness" << "Profile File"; }
    QStringList parmDescs() const { return QStringList() 
      << "Stack" << "Store" << "X Step" << "Y Step" << "Z Step" << "Brightness" << "Profile File"; }
    QStringList parmDefaults() const { return QStringList() 
      << "-1" << "Current" << "1.0" << "1.0" << "1.0" << "1.0" << ""; }
    ParmChoiceMap parmChoice() const
    {
      ParmChoiceMap map;
      map[1] << "Main" << "Work" << "Current";
      return map;
    }
    QIcon icon() const { return QIcon(":/images/open.png"); }
  
  protected slots:
    void SaveProfile();
    void LoadProfile(QString filename = QString());
    void AddFilesSlot();
    void AddFilesSlot(const QStringList& files);
    void RemoveFilesSlot();
    void FilterFilesSlot();
    void SortAscendingSlot(bool val = true);
    void SortDescendingSlot();
  
  protected:
    bool LoadProfile(QString filename, Point3u& size, Point3f& step, float& brightness, QStringList& files);
    void setImageSize(Point3u size);
    void setImageSize(uint x, uint y, uint z) {
      setImageSize(Point3u(x, y, z));
    }
    void setImageResolution(Point3f step);
    void setImageResolution(float x, float y, float z) {
      setImageResolution(Point3f(x, y, z));
    }
    Point3f imageResolution();
  
    QString txtFile;
    QDialog* dlg;
    Ui_LoadStackDialog* ui;
    QString loadedFile;
    Point3u imageSize;
    QStringList imageFiles;
  };
  
  /**
   * \class StackOpen SystemProcessLoad.hpp <SystemProcessLoad.hpp>
   *
   * Open a stack in either the MGXS or INRIA formats.
   *
   * \ingroup StackProcess
   */
  class mgx_EXPORT StackOpen : public StackProcess {
  public:
    StackOpen(const StackProcess& proc) : Process(proc), StackProcess(proc) {}
  
    bool operator()(const QStringList& parms);
    /**
     * Open the image
     * \param stack Stack that will contain the image
     * \param store Store that will contain the image
     * \param filename File containing the image.
     */
    bool operator()(Stack* stack, Store* store, QString filename);
    QString folder() const { return "System"; }
    QString name() const { return "Open"; }
    QString description() const { return "Open a stack from a known 3D image format"; }
    QStringList parmNames() const { return QStringList() << "Filename" << "Store" << "Stack number"; }
    QStringList parmDescs() const { return QStringList() << "Filename" << "Store" << "Stack number"; }
    QStringList parmDefaults() const { return QStringList() << "" << "Main" << "0"; }
    ParmChoiceMap parmChoice() const
    {
      ParmChoiceMap map;
      map[1] = storeChoice();
      return map;
    }
    bool initialize(QStringList& parms, QWidget* parent);
    QIcon icon() const { return QIcon(":/images/open.png"); }
  
    bool loadMGXS_1_3(QIODevice& file, Stack* stack, Store* store);
    bool loadMGXS_1_2(QIODevice& file, Stack* stack, Store* store);
    bool loadMGXS_1_1(QIODevice& file, Stack* stack, Store* store);
    bool loadMGXS_1_0(QIODevice& file, Stack* stack, Store* store);
    bool loadMGXS_0(QIODevice& file, Stack* stack, Store* store);
    void centerImage();
  };
  
  /**
   * \class MeshLoad SystemProcessLoad.hpp <SystemProcessLoad.hpp>
   *
   * Load a MGXM mesh file.
   *
   * \ingroup MeshProcess
   */
  class mgx_EXPORT MeshLoad : public QObject, public MeshProcess {
    Q_OBJECT
  public:
    MeshLoad(const MeshProcess& proc) : Process(proc), QObject(), MeshProcess(proc) {}
  
    bool operator()(const QStringList& parms);
    /**
     * Load a MGXM file
     * \param mesh Mesh that will contain the data
     * \param filename File containing the mesh.
     * \param scale If true, the mesh will be scaled according to the current
     * scale factor
     * \param transform If true, the mesh will be transformed according to
     * the current transformation frame
     * \param add If true, the mesh will be added to the current one instead
     * of replacing it.
     */
    bool operator()(Mesh* mesh, QString filename, bool transform, bool add);
    QString folder() const { return "System"; }
    QString name() const { return "Load"; }
    QString description() const { return "Load a mesh from one of the known formats."; }
    QStringList parmNames() const { return QStringList() 
      << "Filename" << "Transform" << "Add" << "Stack number"; }
    QStringList parmDescs() const { return QStringList() 
      << "Filename" << "Transform" << "Add" << "Stack number"; }
    QStringList parmDefaults() const { return QStringList() 
      << "" << "no" << "no" << "0"; }
    ParmChoiceMap parmChoice() const
    {
      ParmChoiceMap map;
      map[1] = map[2] = booleanChoice();
      return map;
    }
    bool initialize(QStringList& parms, QWidget* parent);
    QIcon icon() const { return QIcon(":/images/open.png"); }
  
    bool loadMGXM_0(QIODevice& file, Mesh* mesh, bool scale, bool transform, bool add, bool has_color = true);
    bool loadMGXM_1_0(QIODevice& file, Mesh* mesh, bool& scale, bool& transform, bool add, bool has_color = true);
    bool loadMGXM_1_1(QIODevice& file, Mesh* mesh, bool& scale, bool& transform, bool add, bool has_color = true);
    bool loadMGXM_1_2(QIODevice& file, Mesh* mesh, bool& scale, bool& transform, bool add, bool has_color = true);
    bool loadMGXM_1_3(QIODevice& file, Mesh* mesh, bool& scale, bool& transform, bool add);
  
    void findSignalBounds(Mesh* mesh);
  
  protected slots:
    void selectMeshFile();
  
  protected:
    void setMeshFile(const QString& filename);
  
    QDialog* dlg;
    Ui_LoadMeshDialog* ui;
  };
  
  /**
   * \class MeshImport SystemProcessLoad.hpp <SystemProcessLoad.hpp>
   *
   * Import a mesh from a file of another format
   *
   * \ingroup MeshProcess
   */
  class mgx_EXPORT MeshImport : public QObject, public MeshProcess {
    Q_OBJECT
  
  public:
    MeshImport(const MeshProcess& proc) : Process(proc), QObject(), MeshProcess(proc) {}
  
    bool operator()(const QStringList& parms);
    /**
     * Import a mesh
     * \param mesh Mesh that will contain the data
     * \param filename File describing the mesh
     * \param type Type of the file. Must be one of 'Text', 'Cells', 'Keyence' or 'MeshEdit', or empty (i.e.
     * auto-detect)
     * \param scale If true, the mesh will be scaled
     * \param transform If true, the mesh will be transformed
     * \param add If true, the mesh won't be cleared before loading the file
     *
     * \sa MeshOpen
     */
    bool initialize(QStringList& parms, QWidget* parent);

    bool operator()(Mesh* mesh, QString filename, QString type, bool scale, bool transform, bool add);

    QString folder() const { return "System"; }
    QString name() const { return "Import"; }
    QString description() const { return "Import a mesh from one of the known formats."; }
    QStringList parmNames() const { return QStringList() 
      << "Filename" << "Kind" << "Scale" << "Transform" << "Add" << "Stack number"; }
    QStringList parmDescs() const { return QStringList() 
      << "Filename" << "Kind" << "Scale" << "Transform" << "Add" << "Stack number"; }
    QStringList parmDefaults() const { return QStringList() 
      << "" << "PLY" << "yes" << "no" << "no" << "0"; }
    ParmChoiceMap parmChoice() const
    {
      ParmChoiceMap map;
      map[1] = QStringList() << "VTK Mesh" << "Text" << "Cells" << "Keyence" << "MeshEdit";
      map[2] = map[3] = map[4] = booleanChoice();
      return map;
    }
    QIcon icon() const { return QIcon(":/images/open.png"); }
  
  protected slots:
    void selectMeshFile();
    void selectMeshType(const QString& type);
  
  protected:
    void setMeshFile(const QString& filename, const QString& type = QString());
    QString properFile(QString filename, const QString& type) const;
  
    bool loadText(Mesh* mesh, const QString& filename, bool scale, bool transform, bool add);
    bool loadCells(Mesh* mesh, const QString& filename, bool scale, bool transform, bool add);
    bool loadKeyence(Mesh* mesh, const QString& filename, bool scale, bool transform, bool add);
    bool loadMeshEdit(Mesh* mesh, const QString& filename, bool scale, bool transform, bool add);
    bool loadMeshVTK(Mesh* mesh, const QString& filename, bool& scale, bool transform, bool add);
    bool loadMeshOBJ(Mesh* mesh, const QString& filename, bool scale, bool transform, bool add);
    bool loadMeshPLY(Mesh* mesh, const QString& filename, bool scale, bool transform, bool add);
  
    QDialog* dlg;
    Ui_ImportMeshDialog* ui;
  };
  
  /**
   * \class LoadAllData SystemProcessLoad.hpp <SystemProcessLoad.hpp>
   *
   * Load all the files as specified in the various objects (filenames).
   *
   * Note that currently, it will load only the main store of the stacks as
   * only they can have any data.
   *
   * \ingroup GlobalProcess
   */
  class mgx_EXPORT LoadAllData : public GlobalProcess {
  public:
    LoadAllData(const GlobalProcess& proc) : Process(proc), GlobalProcess(proc) {}
  
    bool operator()(const QStringList&);
    bool operator()();
  
    QString folder() const { return "System"; }
    QString name() const { return "Load All"; }
    QString description() const { return 
      "Load the data for all existing objects, using the filename and properties set in them."; }
    QStringList parmNames() const { return QStringList(); }
    QStringList parmDescs() const { return QStringList(); }
    QIcon icon() const { return QIcon(":/images/open.png"); }
  
  protected:
    bool loadStore(Stack* stack, Store* store, QStringList& errors);
  };
  
  /**
   * \class LoadViewFile SystemProcessLoad.hpp <SystemProcessLoad.hpp>
   * Load a view file and all the associated files.
   *
   * \ingroup GlobalProcess
   */
  class mgx_EXPORT LoadViewFile : public GlobalProcess {
  public:
    LoadViewFile(const GlobalProcess& proc) : Process(proc), GlobalProcess(proc) {}
  
    bool operator()(const QStringList& parms);
    /**
     * Load the file \c filename
     */
    bool operator()(QString filename);
  
    bool initialize(QStringList& parms, QWidget* parent);
  
    QString folder() const { return "System"; }
    QString name() const { return "Load View"; }
    QString description() const { return 
      "Load a view file and set all the fields and interface. Does not load the data though."; }
    QStringList parmNames() const { return QStringList() << "Filename"; }
    QStringList parmDescs() const { return QStringList() << "Filename"; }
    QStringList parmDefaults() const { return QStringList() << ""; }
    QIcon icon() const { return QIcon(":/images/open.png"); }
  };
  
  /**
   * \class ResetMeshProcess SystemProcessLoad.hpp <SystemProcessLoad.hpp>
   *
   * Reset a mesh
   *
   * \ingroup MeshProcess
   */
  class mgx_EXPORT ResetMeshProcess : public MeshProcess {
  public:
    ResetMeshProcess(const MeshProcess& proc) : Process(proc), MeshProcess(proc) {}
  
    bool operator()(const QStringList& parms);
    bool operator()(Mesh* m);
  
    QString folder() const { return "System"; }
    QString name() const { return "Reset"; }
    QString description() const { return "Reset a mesh, -1 for current."; }
    QStringList parmNames() const { return QStringList() << "Mesh"; }
    QStringList parmDescs() const { return QStringList() << "Mesh"; }
    QStringList parmDefaults() const { return QStringList() << "-1"; }
    QIcon icon() const { return QIcon(":/images/ClearStack.png"); }
  };

  /**
   * \class SaveParents MeshProcessLineage.hpp <MeshProcessLineage.hpp>
   *
   * Save mapping from labels to parents
   *
   * \ingroup MeshProcess
   */
  class mgx_EXPORT SaveParents : public MeshProcess {
  public:
    SaveParents(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool initialize(QStringList& parms, QWidget* parent);
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_ANY))
        return false;
      Mesh* m = currentMesh();
      return (*this)(m, parms[0]);
    }
  
    bool operator()(Mesh* mesh, const QString& filename);
  
    QString folder() const { return "Lineage Tracking"; }
    QString name() const { return "Save Parents"; }
    QString description() const { return "Save map of labels to parents labels to a file"; }
    QStringList parmNames() const { return QStringList() << "Filename"; }
    QStringList parmDescs() const { return QStringList() << "Path to spreadsheet file."; }
    QStringList parmDefaults() const { return QStringList() << ""; }
    QIcon icon() const { return QIcon(":/images/ParentsSave.png"); }
  };
  
  /**
   * \class LoadParents MeshProcessLineage.hpp <MeshProcessLineage.hpp>
   *
   * Load mapping from labels to parents
   *
   * \ingroup MeshProcess
   */
  class mgx_EXPORT LoadParents : public MeshProcess {
  public:
    LoadParents(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool initialize(QStringList& parms, QWidget* parent);
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_ANY))
        return false;
      Mesh* m = currentMesh();
      bool res = (*this)(m, parms[0]);
      if(res)
        m->showParent();
      return res;
    }
  
    bool operator()(Mesh* mesh, const QString& filename);
  
    QString folder() const { return "Lineage Tracking"; }
    QString name() const { return "Load Parents"; }
    QString description() const { return "Load map of labels to parents from a file"; }
    QStringList parmNames() const { return QStringList() << "Filename"; }
    QStringList parmDescs() const { return QStringList() << "Path to spreadsheet file."; }
    QStringList parmDefaults() const { return QStringList() << ""; }
    QIcon icon() const { return QIcon(":/images/ParentsOpen.png"); }
  };
  
  /**
   * \class ResetParents MeshProcessLineage.hpp <MeshProcessLineage.hpp>
   *
   * Clear label to parent mapping
   *
   * \ingroup MeshProcess
   */
  class mgx_EXPORT ResetParents : public MeshProcess {
  public:
    ResetParents(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& )
    {
      if(!checkState().mesh(MESH_ANY))
        return false;
      Mesh* m = currentMesh();
      return (*this)(m);
    }
  
    bool operator()(Mesh* mesh);
  
    QString folder() const { return "Lineage Tracking"; }
    QString name() const { return "Reset Parents"; }
    QString description() const { return "Clear mapping from parents to labels"; }
    QStringList parmNames() const { return QStringList(); }
    QStringList parmDescs() const { return QStringList(); }
    QIcon icon() const { return QIcon(":/images/ParentsClear.png"); }
  };

}}

#endif
