//
// This file is part of MorphoGraphX - http://www.MorphoGraphX.org
// Copyright (C) 2012-2015 Richard S. Smith and collaborators.
//
// If you use MorphoGraphX in your work, please cite:
//   http://dx.doi.org/10.7554/eLife.05864
//
// MorphoGraphX is free software, and is licensed under under the terms of the 
// GNU General (GPL) Public License version 2.0, http://www.gnu.org/licenses.
// 
#ifndef VERTEX_HPP
#define VERTEX_HPP

/**
 * \file Vertex.hpp
 *
 * This file contain the definition of the graph::Vertex class.
 * For now, the graph::Vertex class can only be used with the graph::VVGraph
 * class.
 */

#include <Config.hpp>

#include <Mangling.hpp>
#include <Thrust.hpp>
#include <UnorderedMap.hpp>

#include <iostream>
#include <map>
#include <memory>
#ifndef _MSC_VER
#  include <stdint.h>
#endif
#include <typeinfo>
#include <utility>

namespace mgx {
namespace graph {
/**
 * Type of the identifier of a vertex
 */
typedef intptr_t vertex_identity_t;

/**
 * Number used to enumerate the vertices of all types.
 */
extern mgx_EXPORT size_t vertex_counter;

template <typename VertexContent, typename Alloc> class WeakVertex;

/**
 * \class Vertex vertex.h <graph/vertex.h>
 * Vertex of a vv graph.
 *
 * The vertexes handle their associated data using a reference counting scheme.
 * As such, they can be used as smart pointers. They are also comparable
 * (<,>,==,!=), which allow for use in any sorted structure and hashable for
 * use in any hash table-based structure.
 *
 * They also all have a unique identifier. This identifier can be used to
 * retrieve a weak reference on the data.
 *
 * \warning A weak reference won't keep the data alive! So be careful when you
 * use those. You can test if a vertex hold a weak reference using the
 * Vertex<VertexContent>::isWeakRef() method.
 */
template <typename VertexContent, typename Alloc> class Vertex {
  friend class WeakVertex<VertexContent, Alloc>;

protected:
  /**
   * Type of the reference counted content
   */
  struct CountedContent : public VertexContent {
private:
    CountedContent()
      : VertexContent()
      , num(vertex_counter++)
    {
      count = 1u;       // To be used with tbb::atomic
    }

    CountedContent(const CountedContent& copy)
      : VertexContent(copy)
      , count(copy.count)
      , num(vertex_counter++)
    {
    }

    CountedContent(const VertexContent& copy)
      : VertexContent(copy)
      , num(vertex_counter++)
    {
      count = 1u;       // To be used with tbb::atomic
    }

public:
    typedef typename Alloc::template rebind<CountedContent>::other alloc_t;
    static alloc_t alloc;

    static CountedContent* New()
    {
      CountedContent* cc = alloc.allocate(1);
      ::new (cc) CountedContent();
      return cc;
    }

    static CountedContent* New(const VertexContent& copy)
    {
      CountedContent* cc = alloc.allocate(1);
      ::new (cc) CountedContent(copy);
      return cc;
    }

    static void Delete(CountedContent* cc)
    {
      cc->~CountedContent();
      alloc.deallocate(cc, 1);
    }

    /**
     * Reference count on the vertex
     */
    unsigned int count;

    size_t num;
  };

public:
  /**
   * Type of the identifier of the vertex
   */
  typedef vertex_identity_t identity_t;

  /**
   * Type of the content of the vertex
   */
  typedef VertexContent content_t;

  typedef WeakVertex<VertexContent, Alloc> weak_ref_t;

  /**
   * Type of the equivalent pointer
   */
  typedef VertexContent* pointer;

  /**
   * Creates a new vertex with a new content.
   *
   * Example:
   * \code
   *   struct VertexContent { int a; }
   *   // ...
   *   Vertex<VertexContent> v;
   *   v->a = 10;
   * \endcode
   */
  Vertex();

  /**
   * Creates a reference on the vertex of identifier \c id. If \c id
   * is 0, creates a null vertex.
   *
   * \param[in] id Label of the vertex to retrieve.
   *
   * \warning This function is very unsafe if used with anything but 0 as
   * an identifier.
   *
   * Example:
   * \code
   *   typedef Vertex<VertexContent> vertex;
   *   vertex v;
   *   vertex::identity_t i = v.id();
   *   vertex v1(i);
   *   assert(v1 == v);
   * \endcode
   */
  explicit Vertex(identity_t id);

  /**
   * Copy a vertex.
   *
   * The data is not copied. The quality of the copy (i.e. weak/strong
   * reference) is the same as the copied.
   */
  Vertex(const Vertex& copy);

  /**
   * Construct a strong reference from a weak one
   */
  explicit Vertex(const weak_ref_t& w);

  /**
   * Desctructor
   */
  ~Vertex();

  /**
   * Access to the data.
   *
   * \warning Do not try to access the data of the null vertex.
   */
  VertexContent* operator->() const {
    return content;
  }
  /**
   * Access to the data.
   *
   * \warning Do not try to access the data of the null vertex.
   */
  VertexContent& operator*() const {
    return *content;
  }

  /**
   * Change the vertex held by the current object.
   *
   * The data is never modified by this operation. If you wish to copy the data
   * of a vertex v1 into a vertex v2 use:
   * \code
   * *v2 = *v1;
   * \endcode
   */
  Vertex& operator=(const Vertex& other);

  Vertex& operator=(const identity_t& id);

  Vertex& operator=(const weak_ref_t& other);

  Vertex& operator=(const VertexContent* value);

  /**
   * Comparison operators.
   *
   * \note the comparisons work on the identifiers, not the contents!
   */
  bool operator==(const Vertex& other) const {
    return id() == other.id();
  }
  /**
   * Comparison operators.
   *
   * \note the comparisons work on the identifiers, not the contents!
   */
  bool operator!=(const Vertex& other) const {
    return id() != other.id();
  }
  /**
   * Comparison operators.
   *
   * \note the comparisons work on the identifiers, not the contents!
   */
  bool operator>(const Vertex& other) const {
    return id() > other.id();
  }
  /**
   * Comparison operators.
   *
   * \note the comparisons work on the identifiers, not the contents!
   */
  bool operator<(const Vertex& other) const {
    return id() < other.id();
  }
  /**
   * Comparison operators.
   *
   * \note the comparisons work on the identifiers, not the contents!
   */
  bool operator>=(const Vertex& other) const {
    return id() >= other.id();
  }
  /**
   * Comparison operators.
   *
   * \note the comparisons work on the identifiers, not the contents!
   */
  bool operator<=(const Vertex& other) const {
    return id() <= other.id();
  }

  /**
   * Test if a vertex is a null vertex.
   */
  bool isNull() const {
    return content == 0;
  }

  /**
   * Return the identifier of a vertex.
   */
  identity_t id() const {
    return (identity_t)content;
  }

  /**
   * Return a number unique to each vertex, globally.
   */
  size_t num() const {
    return content->num;
  }

  /**
   * Convert a vertex to \c true if it is not null.
   */
  operator bool() const { return content; }

  /**
   * Return \c true if the current object hold a weak reference on a vertex.
   */
  bool isWeakRef() const {
    return false;
  }

  /**
   * Construct a weak reference on the current vertex
   */
  weak_ref_t weakRef() const;

  /**
   * Null vertex.
   *
   * Useful to return a constant reference on a vertex all the time.
   *
   * Example:
   * \code
   * const vertex& fct(bool cond)
   * {
   *   static vertex a;
   *   if(cond)
   *     return a;
   *    return vertex::null;
   * }
   * \endcode
   */
  static Vertex null;

  /**
   * Serialization method
   *
   * \warning You need to include <storage/graph.h> to use this
   * serialization method
   */
  // bool serialize(storage::VVEStorage&);

  unsigned int count() const
  {
    if(content)
      return content->count;
    return 0;
  }

protected:
  /**
   * Content of the vertex
   *
   * This member is mutable to allow for modification of constant
   * references. This is useful as no operation on the vertex depend on
   * this.
   */
  mutable CountedContent* content;

  /**
   * Release the current pointer
   */
  void release();

  /**
   * Acquire the current pointer
   */
  void acquire();
};

template <typename VertexContent, typename Alloc>
typename Vertex<VertexContent, Alloc>::CountedContent::alloc_t Vertex<VertexContent, Alloc>::CountedContent::alloc;

template <typename VertexContent, typename Alloc> Vertex<VertexContent, Alloc> Vertex<VertexContent, Alloc>::null(0);

template <typename VertexContent, typename Alloc> class WeakVertex : public Vertex<VertexContent, Alloc> {
  typedef typename Vertex<VertexContent, Alloc>::CountedContent CountedContent;

public:
  /**
   * Type of the identifier of the vertex
   */
  typedef vertex_identity_t identity_t;

  /**
   * Type of the content of the vertex
   */
  typedef VertexContent content_t;

  /**
   * Strong reference corresponding to the weak one
   */
  typedef Vertex<VertexContent, Alloc> strong_ref;

  /**
   * Type of the equivalent pointer
   */
  typedef VertexContent* pointer;

  /**
   * Construct an empty weak vertex.
   *
   * At the difference of normal vertices, constructing a weak vertex creates
   * a null one.
   */
  WeakVertex()
    : Vertex<VertexContent, Alloc>(0)
  {
  }

  /**
   * Construct a weak reference from a strong one
   */
  WeakVertex(const strong_ref& v)
    : Vertex<VertexContent, Alloc>(0)
  {
    this->content = v.content;
  }

  /**
   * Copy constructor
   */
  WeakVertex(const WeakVertex& copy)
    : Vertex<VertexContent, Alloc>(0)
  {
    this->content = copy.content;
  }

  /**
   * Construct a weak reference from an id
   */
  explicit WeakVertex(const identity_t& id)
    : Vertex<VertexContent, Alloc>(0)
  {
    this->content = reinterpret_cast<CountedContent*>(id);
    if(this->content and this->content->count == 0)
      this->content = 0;
  }

  /**
   * Set the content of the weak reference
   */
  WeakVertex& operator=(const WeakVertex& other)
  {
    this->content = other.content;
    return *this;
  }

  WeakVertex& operator=(const identity_t& id)
  {
    this->content = reinterpret_cast<CountedContent*>(id);
    return *this;
  }

  WeakVertex& operator=(const strong_ref& other)
  {
    this->content = other.content;
    return *this;
  }

  ~WeakVertex() {
    this->content = 0;
  }

  bool isNull() const {
    return this->content == 0 or this->content->count == 0;
  }

protected:
};

template <typename VertexContent, typename Alloc> Vertex<VertexContent, Alloc>::~Vertex() {
  this->release();
}

template <typename VertexContent, typename Alloc>
Vertex<VertexContent, Alloc>::Vertex()
  : content(0)
{
  content = CountedContent::New();
}

template <typename VertexContent, typename Alloc>
Vertex<VertexContent, Alloc>::Vertex(identity_t id)
  : content(reinterpret_cast<CountedContent*>(id))
{
  acquire();
}

//  template <typename VertexContent, typename Alloc>
//    Vertex<VertexContent,Alloc>::Vertex( VertexContent* data )
//    : content(0)
//  {
//    CountedContent *cdata = dynamic_cast<CountedContent*>(data);
//    if(cdata != 0)
//    {
//      content = cdata;
//      ++(content->count);
//    }
//    else
//    {
//      content = CountedContent::New();
//      *(VertexContent*)content = *data;
//    }
//  }

template <typename VertexContent, typename Alloc>
Vertex<VertexContent, Alloc>::Vertex(const typename Vertex<VertexContent, Alloc>::weak_ref_t& copy)
  : content(copy.content)
{
  acquire();
}

template <typename VertexContent, typename Alloc>
Vertex<VertexContent, Alloc>::Vertex(const Vertex& copy)
  : content(copy.content)
{
  acquire();
}

template <typename VertexContent, typename Alloc>
Vertex<VertexContent, Alloc>& Vertex<VertexContent, Alloc>::operator=(const identity_t& id)
{
  if((identity_t)content == id)
    return *this;
  this->release();
  content = reinterpret_cast<CountedContent*>(id);
  acquire();
  return *this;
}

template <typename VertexContent, typename Alloc>
Vertex<VertexContent, Alloc>& Vertex<VertexContent, Alloc>::operator=(const VertexContent* id)
{
  const CountedContent* cid = dynamic_cast<const CountedContent*>(id);
  if(cid != 0)
    return *this = (const identity_t&)cid;
  else {
    return *this = (const identity_t&)(*(CountedContent::New(*id)));
  }
}

template <typename VertexContent, typename Alloc>
Vertex<VertexContent, Alloc>& Vertex<VertexContent, Alloc>::operator=(const weak_ref_t& copy)
{
  if(content == copy.content)
    return *this;
  this->release();
  content = copy.content;
  acquire();
  return *this;
}

template <typename VertexContent, typename Alloc> void Vertex<VertexContent, Alloc>::acquire()
{
  if(content) {
    if(content->count == 0)
      content = 0;
    else
#pragma omp atomic
      ++(content->count);
  }
}

template <typename VertexContent, typename Alloc> void Vertex<VertexContent, Alloc>::release()
{
  if(content) {
#pragma omp atomic
    --(content->count);
    if(content->count == 0) {
      CountedContent::Delete(content);
    }
  }
}

template <typename VertexContent, typename Alloc>
typename Vertex<VertexContent, Alloc>::weak_ref_t Vertex<VertexContent, Alloc>::weakRef() const
{
  return weak_ref_t(*this);
}

template <typename VertexContent, typename Alloc>
Vertex<VertexContent, Alloc>& Vertex<VertexContent, Alloc>::operator=(const Vertex& other)
{
  if(content == other.content) {
    return *this;
  } else
    this->release();
  content = other.content;
  acquire();
  return *this;
}

template <typename VertexContent, typename Alloc, typename charT>
std::basic_ostream<charT>& operator<<(std::basic_ostream<charT>& ss, const Vertex<VertexContent, Alloc>& v)
{
  ss << "Vertex<" << util::demangle(typeid(VertexContent).name()) << ">(" << v.id() << ")";
  return ss;
}
} // namespace graph
} // namespace mgx

namespace std {
#ifdef HASH_NEED_TR1
namespace tr1 {
#endif
template <typename VertexContent, typename Alloc>
struct hash<mgx::graph::Vertex<VertexContent, Alloc> > {
  size_t operator()(const mgx::graph::Vertex<VertexContent, Alloc>& v) const
  {
    return ((size_t)v.id()) >> 2;         // Shift to consider memory alignment!
  }
};

template <typename VertexContent, typename Alloc>
struct hash<mgx::graph::WeakVertex<VertexContent, Alloc> > {
  size_t operator()(const mgx::graph::WeakVertex<VertexContent, Alloc>& v) const
  {
    return ((size_t)v.id()) >> 2;     // Shift to consider memory alignment!
  }
};
#ifdef HASH_NEED_TR1
} // namespace tr1
#endif
} // namespace std

#endif // VERTEX_HPP
