//
// This file is part of MorphoGraphX - http://www.MorphoGraphX.org
// Copyright (C) 2012-2015 Richard S. Smith and collaborators.
//
// If you use MorphoGraphX in your work, please cite:
//   http://dx.doi.org/10.7554/eLife.05864
//
// MorphoGraphX is free software, and is licensed under under the terms of the 
// GNU General (GPL) Public License version 2.0, http://www.gnu.org/licenses.
// 
#ifndef MergeStacks_HPP
#define MergeStacks_HPP

#include <Process.hpp>

namespace mgx {
namespace process {

///\addtogroup StackProcess
///@{
/**
 * \class AlignCanvas MergeStacks.hpp <MergeStacks.hpp>
 *
 * Align two stacks' canvas. Expand the canvas of the first stack so the
 * second one can be drawn onto it. If the 'change both stack' option is
 * selected, then all the stacks are extended and made aligned with the axis
 * of the reference system. Otherwise, the work store of the current stack is
 * replaced with the projection of the other stack onto the extended canvas,
 * keeping the resolution of the current stack.
 */
class AlignCanvas : public StackProcess {
public:
  AlignCanvas(const StackProcess& process)
    : Process(process)
    , StackProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(not checkState().store(STORE_NON_EMPTY | STORE_VISIBLE, 0).store(STORE_NON_EMPTY | STORE_VISIBLE, 1))
      return false;
    Stack* current = currentStack();
    Stack* other = stack(1 - current->id());
    if(current->currentStore()->labels() != other->currentStore()->labels())
      return setErrorMessage("Error, one stack is labeled and the other is not.");
    bool change_both = stringToBool(parms[0]);
    bool result = operator()(current, other, change_both);
    if(not change_both and result)
      current->work()->show();
    return result;
  }

  bool operator()(Stack* target, Stack* other, bool change_both);

  QString folder() const {
    return "Multi-stack";
  }
  QString name() const {
    return "Align Canvas";
  }
  QString description() const
  {
    return "Align two stacks' canvas. Expand the canvas of the first stack so the\n"
           "second one can be drawn onto it. If the 'change both stack' option is \n"
           "selected, then all the stacks are extended and made aligned with the \n"
           "axis of the reference system. Otherwise, the work store of the current\n"
           "stack is replaced with the projection of the other stack onto the\n"
           "extended canvas, keeping the resolution of the current stack.\n";
  }
  QStringList parmNames() const {
    return QStringList() << "Change both stacks";
  }
  QStringList parmDescs() const {
    return QStringList() << "Extend the non-active stacks canvas as well";
  }
  QStringList parmDefaults() const {
    return QStringList() << "True";
  }
  ParmChoiceMap parmChoice() const
  {
    ParmChoiceMap map;
    map[0] = booleanChoice();
    return map;
  }
  QIcon icon() const {
    return QIcon(":/images/AlignCanvas.png");
  }

  bool projectGlobal(Stack* s1, Stack* s2);
  bool projectOnStack(Stack* target, const Stack* to_project);
};

/**
 * \class CombineStacks MergeStacks.hpp <MergeStacks.hpp>
 *
 * Combine the values of the main and work store onto the work store.
 */
class CombineStacks : public StackProcess {
public:
  CombineStacks(const StackProcess& process)
    : Process(process)
    , StackProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(not checkState().store(STORE_NON_EMPTY | STORE_VISIBLE | STORE_MAIN).store(STORE_NON_EMPTY | STORE_VISIBLE
                                                                                  | STORE_WORK))
      return false;
    Stack* current = currentStack();
    if(current->main()->labels() != current->work()->labels())
      return setErrorMessage("Error, one store is labeled and the other is not.");
    return operator()(current, parms[0]);
  }

  bool operator()(Stack* target, QString method);

  QString folder() const {
    return "Multi-stack";
  }
  QString name() const {
    return "Combine Stacks";
  }
  QString description() const {
    return "Combine the values of the main and work store onto the work store.";
  }
  QStringList parmNames() const {
    return QStringList() << "Method";
  }
  QStringList parmDescs() const {
    return QStringList() << "Method";
  }
  QStringList parmDefaults() const {
    return QStringList() << "Max";
  }
  ParmChoiceMap parmChoice() const
  {
    ParmChoiceMap map;
    map[0] = QStringList() << "Max"
                           << "Min"
                           << "Average"
                           << "Product"
                           << "Add"
                           << "Subtract";
    return map;
  }
  QIcon icon() const {
    return QIcon(":/images/CombineStacks.png");
  }
};

/**
 * \class MergeStacks MergeStacks.hpp <MergeStacks.hpp>
 *
 * Merge the main store of the current stack with the current store of the other one.
 * The current stack will be aligned with the other before the stores being combines.
 * The method argument is simply passed to the Combine_Stacks process.
 */
class MergeStacks : public StackProcess {
public:
  MergeStacks(const StackProcess& process)
    : Process(process)
    , StackProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(not checkState().store(STORE_NON_EMPTY | STORE_VISIBLE, 0).store(STORE_NON_EMPTY | STORE_VISIBLE, 1))
      return false;
    Stack* current = currentStack();
    Stack* other = stack(1 - current->id());
    if(current->currentStore() != current->main())
      return setErrorMessage("Error, the current stack must have its main store active.");
    if(current->currentStore()->labels() != other->currentStore()->labels())
      return setErrorMessage("Error, one stack is labeled and the other is not.");
    bool result = operator()(current, other, parms[0]);
    if(result)
      current->work()->show();
    return result;
  }

  bool operator()(Stack* target, const Stack* other, QString method);

  QString folder() const {
    return "Multi-stack";
  }
  QString name() const {
    return "Merge Stacks";
  }
  QString description() const
  {
    return "Merge the main store of the current stack with the current store of the other one.\n"
           "The current stack will be aligned with the other before the stores being combines.\n"
           "The method argument is simply passed to the Combine_Stacks process.";
  }
  QStringList parmNames() const {
    return QStringList() << "Method";
  }
  QStringList parmDescs() const {
    return QStringList() << "Method";
  }
  QStringList parmDefaults() const {
    return QStringList() << "max";
  }
  ParmChoiceMap parmChoice() const
  {
    ParmChoiceMap map;
    map[0] = QStringList() << "max"
                           << "min"
                           << "average"
                           << "product";
    return map;
  }
  QIcon icon() const {
    return QIcon(":/images/MergeStacks.png");
  }
};
///@}
} // namespace process
} // namespace mgx
#endif
