//
// This file is part of MorphoGraphX - http://www.MorphoGraphX.org
// Copyright (C) 2012-2015 Richard S. Smith and collaborators.
//
// If you use MorphoGraphX in your work, please cite:
//   http://dx.doi.org/10.7554/eLife.05864
//
// MorphoGraphX is free software, and is licensed under under the terms of the 
// GNU General (GPL) Public License version 2.0, http://www.gnu.org/licenses.
// 
#ifndef MESH_PROCESSS_CELLAXIS_HPP
#define MESH_PROCESSS_CELLAXIS_HPP

#include <Process.hpp>

#include <MeshProcessCellMesh.hpp>

namespace mgx {
namespace process {
///\addtogroup MeshProcess
///@{
/**
 * \class SaveCellAxis <MeshProcessCellAxis.hpp>
 */
class mgxBase_EXPORT SaveCellAxis : public MeshProcess {
public:
  SaveCellAxis(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool initialize(QStringList& parms, QWidget* parent);

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY))
      return false;
    return (*this)(currentMesh(), parms[0]);
  }

  bool operator()(Mesh* mesh, const QString& filename);

  QString folder() const {
    return "Cell Axis";
  }
  QString name() const {
    return "Cell Axis Save";
  }
  QString description() const {
    return "Save PDG, Fibril Orientations or Curvature to a spreadsheet file.";
  }
  QStringList parmNames() const {
    return QStringList() << "Output File";
  }
  QStringList parmDescs() const {
    return QStringList() << "Path to output file";
  }
  QStringList parmDefaults() const {
    return QStringList() << "";
  }
  QIcon icon() const {
    return QIcon(":/images/CellAxisSave.png");
  }
};

/**
 * \class LoadCellAxis <MeshProcessCellAxis.hpp>
 */
class mgxBase_EXPORT LoadCellAxis : public MeshProcess {
public:
  LoadCellAxis(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool initialize(QStringList& parms, QWidget* parent);

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY))
      return false;
    return (*this)(currentMesh(), parms[0], parms[1]);
  }

  bool operator()(Mesh* mesh, const QString& type, const QString& filename);

  QString folder() const {
    return "Cell Axis";
  }
  QString name() const {
    return "Cell Axis Load";
  }
  QString description() const {
    return "Load cell axis from a spreadsheet file.";
  }
  QStringList parmNames() const
  {
    return QStringList() << "Type"
                         << "Input File";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "Load PDG, Fibril Orientations or Curvature from a spreadsheet file."
                         << "Path to input file. If empty, a browser will open.";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "PDG"
                         << "";
  }
  ParmChoiceMap parmChoice() const
  {
    ParmChoiceMap map;
    map[0] = QStringList() << "PDG"
                           << "fibril"
                           << "curvature";
    return map;
  }
  QIcon icon() const {
    return QIcon(":/images/CellAxisOpen.png");
  }
};

/**
 * \class HideCellAxis <MeshProcessCellAxis.hpp>
 *
 * Hide the cell axis on the mesh
 */
class mgxBase_EXPORT ClearCellAxis : public MeshProcess {
public:
  ClearCellAxis(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList&)
  {
    if(!checkState().mesh(MESH_NON_EMPTY))
      return false;
    return (*this)(currentMesh());
  }

  bool operator()(Mesh* mesh);

  QString folder() const {
    return "Cell Axis";
  }
  QString name() const {
    return "Cell Axis Clear";
  }
  QString description() const {
    return "Remove any cell axis information from the current mesh.";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon(":/images/CellAxisClear.png");
  }
};

/**
 * \class HideCellAxis <MeshProcessCellAxis.hpp>
 *
 * Hide the cell axis on the mesh
 */
class mgxBase_EXPORT HideCellAxis : public MeshProcess {
public:
  HideCellAxis(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& )
  {
    if(!checkState().mesh(MESH_NON_EMPTY))
      return false;
    return (*this)(currentMesh());
  }

  bool operator()(Mesh* mesh);

  QString folder() const {
    return "Cell Axis";
  }
  QString name() const {
    return "Cell Axis Hide";
  }
  QString description() const {
    return "Hide the cell axis on the current mesh.";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon(":/images/CellAxisHide.png");
  }
};
///@}
} // namespace process
} // namespace mgx
#endif
