//
// This file is part of MorphoGraphX - http://www.MorphoGraphX.org
// Copyright (C) 2012-2015 Richard S. Smith and collaborators.
//
// If you use MorphoGraphX in your work, please cite:
//   http://dx.doi.org/10.7554/eLife.05864
//
// MorphoGraphX is free software, and is licensed under under the terms of the 
// GNU General (GPL) Public License version 2.0, http://www.gnu.org/licenses.
// 
#ifndef PRINCIPAL_ORIENTATIONS_HPP
#define PRINCIPAL_ORIENTATIONS_HPP

#include <Process.hpp>

namespace mgx {
namespace process {
///\addtogroup MeshProcess
///@{
/**
 * \class FibrilOrientations <MeshProcessFibril.hpp>
 *
 * Compute principle orientations of lines in the mesh signal. Based on
 * Boudaoud et al., 'FibrilTool, an ImageJ plug-in to quantify fibrillar
 * structures in raw microscopy images', Nature Protocols 2014
 */
class FibrilOrientations : public MeshProcess {
public:
  FibrilOrientations(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY | MESH_SIGNAL))
      return false;
    Mesh* mesh = currentMesh();
    return operator()(mesh, parms[0].toFloat(), parms[1].toFloat());
  }

  bool operator()(Mesh* mesh, float border, float minAreaRatio);

  QString folder() const {
    return "Cell Axis/Fibril Orientations";
  }
  QString name() const {
    return "Compute Fibril Orientations";
  }
  QString description() const
  {
    return "Compute principle orientations of lines in the mesh signal.\n"
           "Based on Boudaoud et al., 'FibrilTool, an ImageJ plug-in to\n"
           "quantify fibrillar structures in raw microscopy images', Nature Protocols 2014";
  }
  QStringList parmNames() const
  {
    return QStringList() << "Border Size"
                         << "Minimum inner area ratio";
  }
  QStringList parmDescs() const
  {
    return QStringList()
           << "Width of cell border that is not taken into account for the computation."
           << "Minimum ratio of inner area (whole cell - border) vs. total area needed for compuation.";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "1.0"
                         << "0.25";
  }
  QIcon icon() const {
    return QIcon(":/images/PrincipalOrientations.png");
  }
};

/**
 * \class DisplayFibrilOrientations <MeshProcessFibril.hpp>
 *
 * Change the representation of the fibril orientation after it has been
 * computed.
 */
class mgxBase_EXPORT DisplayFibrilOrientations : public MeshProcess {
public:
  DisplayFibrilOrientations(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY))
      return false;

    return (*this)(currentMesh(), parms[0], QColor(parms[1]), parms[2].toFloat(), parms[3].toFloat(),
                   parms[4].toFloat(), parms[5].toFloat());
  }

  bool operator()(Mesh* mesh, const QString DisplayHeatMap, const QColor& ColorMax, float AxisLineWidth,
                  float ScaleAxisLength, float AxisOffset, float OrientationThreshold);

  QString folder() const {
    return "Cell Axis/Fibril Orientations";
  }
  QString name() const {
    return "Display Fibril Orientations";
  }
  QString description() const
  {
    return "Display the orientations of fibrils on the image.\n"
           "Only the maximal direction (main orientation) is displayed as a vector.";
  }
  QStringList parmNames() const
  {
    return QStringList() << "Heatmap"
                         << "Line Color"
                         << "Line Width"
                         << "Line Scale"
                         << "Line Offset"
                         << "Threshold";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "Display orientation strength (= MaxDirection/MinDirection - 1) as a colormap."
                         << "Line Color"
                         << "Line Width"
                         << "Length of the vectors = Scale * orientation strength."
                         << "Draw the vector ends a bit tilted up for proper display on surfaces."
                         << "Minimal value of orientation strength required for drawing main direction.";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "none"
                         << "red"
                         << "5.0"
                         << "10.0"
                         << "0.1"
                         << "0.0";
  }
  ParmChoiceMap parmChoice() const
  {
    ParmChoiceMap map;
    map[0] = QStringList() << "none"
                           << "Orientation";
    map[1] = QColor::colorNames();
    return map;
  }
  QIcon icon() const {
    return QIcon(":/images/PrincipalOrientations.png");
  }
};
///@}
} // namespace process
} // namespace mgx

#endif
