//
// This file is part of MorphoGraphX - http://www.MorphoGraphX.org
// Copyright (C) 2012-2015 Richard S. Smith and collaborators.
//
// If you use MorphoGraphX in your work, please cite:
//   http://dx.doi.org/10.7554/eLife.05864
//
// MorphoGraphX is free software, and is licensed under under the terms of the 
// GNU General (GPL) Public License version 2.0, http://www.gnu.org/licenses.
// 
#ifndef MESH_PROCESS_LINEAGE_HPP
#define MESH_PROCESS_LINEAGE_HPP

#include <Process.hpp>

namespace mgx { namespace process {

  ///\addtogroup MeshProcess
  ///@{
  /**
   * \class CorrectParents MeshProcessLineage.hpp <MeshProcessLineage.hpp>
   *
   * Take out non-existing labels from parent map.
   */
  class mgxBase_EXPORT CorrectParents : public MeshProcess {
  public:
    CorrectParents(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& )
    {
      if(!checkState().mesh(MESH_NON_EMPTY, 0).mesh(MESH_NON_EMPTY, 1))
        return false;
      Mesh* mesh1, *mesh2;
      if(currentMesh() == mesh(0)) {
        mesh2 = mesh(0);
        mesh1 = mesh(1);
      } else if(currentMesh() == mesh(1)) {
        mesh2 = mesh(1);
        mesh1 = mesh(0);
      } else
        return false;
  
      bool res = (*this)(mesh1, mesh2);
      return res;
    }
  
    bool operator()(Mesh* mesh1, Mesh* mesh2);
  
    QString folder() const { return "Lineage Tracking"; }
    QString name() const { return "Correct Parents"; }
    QString description() const { return "Take out non-existing labels from parent map."; }
    QStringList parmNames() const { return QStringList(); }
    QStringList parmDescs() const { return QStringList(); }
    QIcon icon() const { return QIcon(":/images/ParentsCheck.png"); }
  };
  
  /**
   * \class HeatMapDaughterCells MeshProcessLineage.hpp <MeshProcessLineage.hpp>
   *
   * Compute the heap map that shows how many daughter cells a parent cell has.
   */
  class HeatMapDaughterCells : public MeshProcess {
  public:
    HeatMapDaughterCells(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& )
    {
      if(!checkState().mesh(MESH_NON_EMPTY | MESH_PARENT)) {
        setErrorMessage("The current mesh must have show parents selected.");
        return false;
      }
      Mesh* m = currentMesh();
  
      bool res = (*this)(m);
      if(res)
        m->showHeat();
      return res;
    }
  
    bool operator()(Mesh* m);
  
    QString folder() const { return "Lineage Tracking"; }
    QString name() const { return "Heat Map Daughter Cells"; }
    QString description() const { return 
      "Compute the heat map that shows how may daughter cells a parent cell has."; }
    QStringList parmNames() const { return QStringList(); }
    QStringList parmDescs() const { return QStringList(); }
    QStringList parmDefaults() const { return QStringList(); }
    QIcon icon() const { return QIcon(":/images/HeatMapDaughterCells.png"); }
  };
  
  /**
   * \class CopyParentsToLabels MeshProcessLineage.hpp <MeshProcessLineage.hpp>
   *
   * Copy parents to labels, and clear parent table.
   */
  class CopyParentsToLabels : public MeshProcess {
  public:
    CopyParentsToLabels(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& )
    {
      if(!checkState().mesh(MESH_NON_EMPTY | MESH_PARENT)) {
        setErrorMessage("The current mesh must have show parents selected.");
        return false;
      }
      Mesh* m = currentMesh();
  
      bool res = (*this)(m);
      if(res)
        m->showLabel();
      return res;
    }
  
    bool operator()(Mesh* m);
  
    QString folder() const { return "Lineage Tracking"; }
    QString name() const { return "Copy Parents to Labels"; }
    QString description() const { return "Copy parents to labels, and clear parent table."; }
    QStringList parmNames() const { return QStringList(); }
    QStringList parmDescs() const { return QStringList(); }
    QStringList parmDefaults() const { return QStringList(); }
    QIcon icon() const { return QIcon(":/images/ParentsCopyToLabel.png"); }
  };
  ///@}
 
}}

#endif
