//
// This file is part of MorphoGraphX - http://www.MorphoGraphX.org
// Copyright (C) 2012-2015 Richard S. Smith and collaborators.
//
// If you use MorphoGraphX in your work, please cite:
//   http://dx.doi.org/10.7554/eLife.05864
//
// MorphoGraphX is free software, and is licensed under under the terms of the 
// GNU General (GPL) Public License version 2.0, http://www.gnu.org/licenses.
// 
#ifndef MESH_PROCESSS_PDG_HPP
#define MESH_PROCESSS_PDG_HPP

#include <Process.hpp>

#include <MeshProcessCellMesh.hpp>

namespace mgx {
namespace process {
///\addtogroup MeshProcess
///@{
/**
 * \class CorrespondanceJunctions MeshProcessPDG.hpp <MeshProcessPDG.hpp>
 */
class CorrespondanceJunctions : public MeshProcess {
public:
  CorrespondanceJunctions(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY, 0).mesh(MESH_NON_EMPTY, 1))
      return false;
    Mesh* mesh1, *mesh2;
    if(currentMesh() == mesh(0)) {
      mesh1 = mesh(0);
      mesh2 = mesh(1);
    } else if(currentMesh() == mesh(1)) {
      mesh2 = mesh(0);
      mesh1 = mesh(1);
    } else
      return false;

    bool res = (*this)(mesh1, mesh2, stringToBool(parms[0]), stringToBool(parms[1]));
    return res;
  }

  bool operator()(Mesh* mesh1, Mesh* mesh2, bool eraseMarginCells, bool ShowVVCorrespondance);

  QString folder() const {
    return "Cell Axis/PDG";
  }
  QString name() const {
    return "Check Correspondence";
  }
  QString description() const
  {
    return "Find matching cell junctions between 2 meshes based on parent labeling.\n"
           "Both meshes are simplified with Make Cells to keep only the cell junctions and centers.\n"
           "SAVE your meshes before running!";
  }
  QStringList parmNames() const
  {
    return QStringList() << "Erase margin cells"
                         << "Show correspondence";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "Take away the cells touching the mesh outer edge while converting to a cellular mesh."
                         << "Draw connecting lines between corresponding junctions.";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "Yes"
                         << "No";
  }
  ParmChoiceMap parmChoice() const
  {
    ParmChoiceMap map;
    map[0] = map[1] = booleanChoice();
    return map;
  }
  QIcon icon() const {
    return QIcon(":/images/CorrespondanceJunctions.png");
  }
};

/**
 * \class GrowthDirections MeshProcessPDG.hpp <MeshProcessPDG.hpp>
 */
class mgxBase_EXPORT GrowthDirections : public MeshProcess {
public:
  GrowthDirections(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }
  bool operator()(const QStringList& )
  {
    if(!checkState().mesh(MESH_NON_EMPTY, 0).mesh(MESH_NON_EMPTY, 1))
      return false;

    Mesh* mesh1, *mesh2;
    if(currentMesh() == mesh(0)) {
      mesh1 = mesh(0);
      mesh2 = mesh(1);
    } else if(currentMesh() == mesh(1)) {
      mesh2 = mesh(0);
      mesh1 = mesh(1);
    } else
      return false;

    bool res = (*this)(mesh1, mesh2);
    return res;
  }

  bool operator()(Mesh* mesh1, Mesh* mesh2);

  QString folder() const {
    return "Cell Axis/PDG";
  }
  QString name() const {
    return "Compute Growth Directions";
  }
  QString description() const
  {
    return "Compute PDGs based on correspondence between junctions.\n"
           "Adapted from Goodall and Green, 'Quantitative Analysis of Surface Growth.'"
           "Botanical Gazette (1986) \n"
           "and Dumais and Kwiatkowska, 'Analysis of surface growth in shoot apices.'"
           "Plant Journal (2002)";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon(":/images/PDG.png");
  }
};

/**
 * \class DisplayPDGs MeshProcessPDG.hpp <MeshProcessPDG.hpp>
 */
class mgxBase_EXPORT DisplayPDGs : public MeshProcess {
public:
  DisplayPDGs(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY))
      return false;

    return (*this)(currentMesh(), parms[0], parms[1], QColor(parms[2]), QColor(parms[3]), parms[4].toFloat(),
                   parms[5].toFloat(), parms[6].toFloat(), parms[7].toFloat());
  }

  bool operator()(Mesh* mesh, const QString DisplayHeatMap, const QString DisplayPDGs, const QColor& ColorExpansion,
                  const QColor& ColorShrinkage, float AxisLineWidth, float ScaleAxisLength, float AxisOffset,
                  float AnisotropyThreshold);

  QString folder() const {
    return "Cell Axis/PDG";
  }
  QString name() const {
    return "Display Growth Directions";
  }
  QString description() const {
    return "Display the principle growth directions";
  }
  QStringList parmNames() const
  {
    return QStringList() << "Heatmap"
                         << "Show Axis"
                         << "Color +"
                         << "Color -"
                         << "Line Width"
                         << "Line Scale"
                         << "Line Offset"
                         << "Threshold";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "Display stretch ratio values in max or min direction as a color map.\n"
                            "stretch ratio = (length after/length before). No deformation means stretch ratio = 1."
                         << "Draw directions as vectors, scaled by strain.\n"
              "strain = (stretch ratio - 1). No deformation means strain = 0. "
                         << "Color used for expansion (strain > 0)"
                         << "Color used for shrinkage (strain < 0)"
                         << "Line Width"
                         << "Length of the vectors = Scale * Strain."
                         << "Draw the vector ends a bit tilted up for proper display on surfaces."
                         << "Minimal value of anisotropy (= stretchMax/StretchMin) required for drawing PDGs.\n"
              "Use a value above 1.";
  }

  QStringList parmDefaults() const
  {
    return QStringList() << "StretchMax"
                         << "both"
                         << "white"
                         << "red"
                         << "2.0"
                         << "10.0"
                         << "0.1"
                         << "0.0";
  }
  ParmChoiceMap parmChoice() const
  {
    ParmChoiceMap map;
    map[0] = QStringList() << "none"
                           << "StretchMax"
                           << "StretchMin"
                           << "Aniso=StretchMax/StretchMin"
                           << "StretchMinXStretchMax";
    map[1] = QStringList() << "both"
                           << "StrainMax"
                           << "StrainMin"
                           << "none";
    map[2] = QColor::colorNames();
    map[3] = QColor::colorNames();
    return map;
  }
  QIcon icon() const {
    return QIcon(":/images/PDG.png");
  }
};
///@}
} // namespace process
} // namespace mgx
#endif
