//
// This file is part of MorphoGraphX - http://www.MorphoGraphX.org
// Copyright (C) 2012-2015 Richard S. Smith and collaborators.
//
// If you use MorphoGraphX in your work, please cite:
//   http://dx.doi.org/10.7554/eLife.05864
//
// MorphoGraphX is free software, and is licensed under under the terms of the 
// GNU General (GPL) Public License version 2.0, http://www.gnu.org/licenses.
// 
#ifndef MESH_PROCESS_SIGNAL_HPP
#define MESH_PROCESS_SIGNAL_HPP

#include <Process.hpp>

#include <Curvature.hpp>

class Ui_LoadHeatMap;

namespace mgx { namespace process {

  typedef std::map<int, HVec3U> IntHVec3UMap;
  
  ///\addtogroup MeshProcess
  ///@{
  /**
   * \class ViewMeshProcess MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * This process is used to change how the mesh is seen. Mostly useful to
   * write scripts taking screenshots.
   */
  class mgxBase_EXPORT ViewMeshProcess : public MeshProcess {
  public:
    ViewMeshProcess(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms);
  
    QString folder() const { return "System"; }
    QString name() const { return "View"; }
    QString description() const { return 
      "Modify how the current mesh is viewed. Useful for scripts."; }
    QStringList parmNames() const { return QStringList() 
      << "Show Surface" << "Surface Type" << "Signal Type" << "Blend" << "Cull"
      << "Show Mesh" << "Mesh View" << "Show Lines" << "Show Points" << "Show Map"
      << "Scale" << "Transform" << "BBox" << "Brightness" << "Opacity";
    }
    QStringList parmDescs() const { return QStringList() 
      << "Draw mesh as a continuous surface."
      << "Nrml: show projected signal, Labels: color triangles according to assigned "
         "labels, Parents: color triangles with parent labels, Heat: color triangles "
         "with computed heat map (e.g. growth)"
      << "Signal: show projected signal, Tex: Use 3D stack as texture, Img: Use 2D "
         "texture (height map)."
      << "Semi-transparent mesh, for example to superimpose to meshes or view the "
         "stack through the mesh."
      << "Color the triangles (with signal or labels) only on the top of the mesh."
      << "Draw triangle edges and nodes"
      << "All: draw all triangles. Border: draw outside edge of the mesh only. "
         "Cells: draw cell outlines only. Selected: draw selected nodes only. "
      << "Show connecting lines between nodes in the mesh."
      << "Show mesh nodes."
      << "Mapping of text on the labels (e.g. label number)"
      << "Change scaling of mesh/stacks, independently in 3 directions (x,y,z). NB: "
      << "a stack saved with 'Scale' turned on will have a modified voxel size, "
         "while saved meshes are unaffected. "
      << "Apply rotation and translation to the mesh/stack."
      << "Display the bounding box (i.e. total size) of a stack."
      << "Brightness of signal, labels or heat"
      << "Opacity of signal, labels or heat";
    }
  
    QStringList parmDefaults() const { return QStringList() 
      << "" << "" << "" << "" << "" << "" << "" << "" << "" << ""
      << "" << "" << "" << "-1" << "-1"; }
  
    ParmChoiceMap parmChoice() const
    {
      QStringList bools = QStringList() << "" << booleanChoice();
      ParmChoiceMap map;
      map[0] = bools;
      map[1] = QStringList() << "" << "Normal" << "Heat" << "Label";
      map[2] = QStringList() << "" << "Signal" << "Texture" << "Image";
      map[3] = map[4] = map[5] = bools;
      map[6] = QStringList() << "" << "All" << "Border" << "Cells" << "Selected";
      map[7] = map[8] = map[9] = map[10] = map[11] = map[12] = bools;
      return map;
    }
    QIcon icon() const { return QIcon(":/images/Palette.png"); }
  };
  
  /**
   * \class ProjectSignal MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * Project stack signal onto the current mesh.
   */
  class mgxBase_EXPORT ProjectSignal : public MeshProcess {
  public:
    ProjectSignal(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().store(STORE_NON_LABEL).mesh(MESH_NON_EMPTY))
        return false;
      return (*this)(currentStack()->currentStore(), currentMesh(), stringToBool(parms[0]), 
        parms[1].toFloat(), parms[2].toFloat(), parms[3].toFloat(), parms[4].toFloat());
    }
  
    bool operator()(const Store* store, Mesh* mesh, bool useAbsSignal, float minDist, 
      float maxDist, float absSignalMin, float absSignalMax);
    void projSignal(const Stack* stack, const HVecUS& data, vertex v, float mindist, float maxdist);
    QString folder() const { return "Signal"; }
    QString name() const { return "Project Signal"; }
    QString description() const { return "Project signal onto mesh, perpendicular to its curved surface."; }
    QStringList parmNames() const { return QStringList() 
      << "Use absolute" << QString("Min Dist (%1)").arg(UM) << QString("Max Dist (%1)").arg(UM)
      << "Min Signal" << "Max Signal";
    }
    QStringList parmDescs() const { return QStringList() 
      << "Use absolute values of signal, instead of normalizing it, useful for signal quantification."
      << "Distance (triangle-voxel) above which the signal is projected."
      << "Maximal distance (triangle-voxel) used for signal projection."
      << "Lower bound of signal value if 'Use absolute' is chosen"
      << "Upper bound of projected signal value.";
    }
    QStringList parmDefaults() const { return QStringList() << "No" << "2.0" << "6.0" << "0.0" << "60000.0"; }
    ParmChoiceMap parmChoice() const
    {
      ParmChoiceMap map;
      map[0] = booleanChoice();
      return map;
    }
    QIcon icon() const { return QIcon(":/images/ProjectColor.png"); }
  };
  
  /**
   * \class SmoothMeshSignal MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * Smooth the mesh signal using local averaging.
   */
  class mgxBase_EXPORT SmoothMeshSignal : public MeshProcess {
  public:
    SmoothMeshSignal(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_NON_EMPTY))
        return false;
      return (*this)(currentMesh(), parms[0].toUInt());
    }
  
    bool operator()(Mesh* mesh, uint passes);
  
    QString folder() const { return "Signal"; }
    QString name() const { return "Smooth Mesh Signal"; }
    QString description() const { return "Averages the signal of each node, based on its immediate neighbors."; }
    QStringList parmNames() const { return QStringList() << "Passes"; }
    QStringList parmDescs() const { return QStringList() << "Number of smoothing iterations."; }
    QStringList parmDefaults() const { return QStringList() << "3"; }
    QIcon icon() const { return QIcon(":/images/SmoothColor.png"); }
  };

  /**
   * \class BrightenMesh MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * Change the brightness of the signal.
   */
  class mgxBase_EXPORT MeshBrightness : public MeshProcess {
  public:
    MeshBrightness(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_NON_EMPTY))
        return false;
      return (*this)(currentMesh(), parms[0].toFloat());
    }
  
    bool operator()(Mesh* mesh, float amount);
  
    QString folder() const { return "Signal"; }
    QString name() const { return "Mesh Brighness"; }
    QString description() const { return "Changes the brightness of the signal on a mesh."; }
    QStringList parmNames() const { return QStringList() << "Amount"; }
    QStringList parmDescs() const { return QStringList() << "Amount to multiply the signal, >1 is brighter"; }
    QStringList parmDefaults() const { return QStringList() << "2.0"; }
    QIcon icon() const { return QIcon(":/images/Brightness.png"); }
  }; 

  /**
   * \class ClearMeshSignal MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * Clear the mesh signal to a defined value.
   */
  class mgxBase_EXPORT ClearMeshSignal : public MeshProcess {
  public:
    ClearMeshSignal(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_NON_EMPTY))
        return false;
      return (*this)(currentMesh(), parms[0].toUInt());
    }
  
    bool operator()(Mesh* mesh, uint value);
  
    QString folder() const { return "Signal"; }
    QString name() const { return "Clear Mesh Signal"; }
    QString description() const { return "Erase the signal on the mesh"; }
    QStringList parmNames() const { return QStringList() << "Value"; }
    QStringList parmDescs() const { return QStringList() << "Assign this signal value to the mesh."; }
    QStringList parmDefaults() const { return QStringList() << "50000"; }
    QIcon icon() const { return QIcon(":/images/ClearSignal.png"); }
  };
  
  /**
   * \class ProjectCurvature MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * Set the mesh signal to be the curvature of the mesh. The process can also
   * output a CSV file with the full curvature tensor for each vertex.
   */
  class mgxBase_EXPORT ProjectCurvature : public MeshProcess {
  public:
    ProjectCurvature(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_NON_EMPTY))
        return false;
      bool ok;
      float neighborhood = parms[2].toFloat(&ok);
      if(not ok)
        return setErrorMessage("Error, parameter 'Neighborhood' must be a number");
      float mincurv = parms[4].toFloat(&ok);
      if(not ok)
        return setErrorMessage("Error, parameter 'Min Curv' must be a number");
      float maxcurv = parms[5].toFloat(&ok);
      if(not ok)
        return setErrorMessage("Error, parameter 'Max Curv' must be a number");
      float percentile = parms[6].toFloat(&ok);
      if(not ok)
        return setErrorMessage("Error, parameter 'Percentile' must be a number");
      return (*this)(currentMesh(), parms[0], parms[1], neighborhood, stringToBool(parms[3]), mincurv, maxcurv,
                     percentile);
    }
  
    bool operator()(Mesh* mesh, QString output, QString type, float neighborhood, bool auto_scale, float mincurv,
                    float maxcurv, int percentile);
  
    QString folder() const { return "Signal"; }
    QString name() const { return "Project Mesh Curvature"; }
    QString description() const { return 
      "Compute curvature at each node of the mesh, for a given neighborhood size. "
      "Curvature values are stored as signal.";
    }
    QStringList parmNames() const { return QStringList() 
      << "Output" << "Type" << QString("Neighborhood (%1)").arg(UM) << "AutoScale"
      << "Min Curv" << "Max Curv" << "Autoscale percentile";
    }
    QStringList parmDescs() const { return QStringList() 
      << "Name of output file, if desired."
      << "Minimal = minCurv, Maximal = maxCurv, Gaussian = maxCurv * minCurv, "
         "SumSquare = maxCurv^2 + minCurv^2, Average = (maxCurv + minCurv)/2, "
         "SignedAverageAbs = sign(max or min) x (abs(maxCurv) + abs(minCurv))/2"
      << QString("Neighborhood (%1)").arg(UM)
      << "Clip max and min signal range according to curvature distribution"
      << "Minimal curvature value displayed"
      << "Maximal curvature value displayed"
      << "Auto-scale signal range based on curvature percentile";
    }
    QStringList parmDefaults() const { return QStringList() 
      << "" << "Gaussian" << "3.0" << "Yes" << "-50.0" << "50.0" << "85"; }
    ParmChoiceMap parmChoice() const
    {
      ParmChoiceMap map;
      map[1] = QStringList() << "Minimal" << "Maximal" << "Gaussian" << "SumSquare" << "Average" << "SignedAverageAbs";
      map[3] = booleanChoice();
      return map;
    }
    QIcon icon() const { return QIcon(":/images/Curvature.png"); }
  };
  
  /**
   * \class MeshGaussianBlur MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * Smooth the mesh signal using a gaussian kernel.
   */
  class mgxBase_EXPORT MeshGaussianBlur : public MeshProcess {
  public:
    MeshGaussianBlur(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_NON_EMPTY))
        return false;
      return (*this)(currentMesh(), parms[0].toFloat());
    }
  
    bool operator()(Mesh* mesh, float radius);
  
    QString folder() const { return "Signal"; }
    QString name() const { return "Gaussian Blur"; }
    QString description() const { return "Apply Gaussian Blur to mesh signal"; }
    QStringList parmNames() const { return QStringList() << QString("Radius (%1)").arg(UM); }
    QStringList parmDescs() const { return QStringList() 
      << "Size of neighborhood used for Gaussian blur. The blur function standard "
         "deviation is given by sigma = radius/2. ";
    }
    QStringList parmDefaults() const { return QStringList() << "2.0"; }
    QIcon icon() const { return QIcon(":/images/Blur.png"); }
  };
  
  /**
   * \class MeshDiffGaussians MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * Calculate a difference of Gaussians on mesh.
   */
  class mgxBase_EXPORT MeshDiffGaussians : public MeshProcess {
  public:
    MeshDiffGaussians(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_NON_EMPTY))
        return false;
      return (*this)(currentMesh(), parms[0].toFloat(), parms[1].toFloat());
    }
  
    bool operator()(Mesh* mesh, float radius1, float radius2);
  
    QString folder() const { return "Signal"; }
    QString name() const { return "Difference of Gaussians"; }
    QString description() const { return "Calculate a difference of Gaussians for the mesh signal"; }
    QStringList parmNames() const { return QStringList() << QString("Radius1 (%1)").arg(UM) << QString("Radius2 (%1)").arg(UM); }
    QStringList parmDescs() const { return QStringList() << "Size of first neighborhood" << "Size of second neighborhood"; }
    QStringList parmDefaults() const { return QStringList() << "1.0" << "5.0"; }
    QIcon icon() const { return QIcon(":/images/Blur.png"); }
  };
  
  /**
   * \class MeshLocalMinima MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * Find the local minima in the current mesh.
   */
  class mgxBase_EXPORT MeshLocalMinima : public MeshProcess {
  public:
    MeshLocalMinima(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_NON_EMPTY))
        return false;
      return (*this)(currentMesh(), parms[0].toFloat());
    }
  
    bool operator()(Mesh* mesh, float radius);
  
    QString folder() const { return "Segmentation"; }
    QString name() const { return "Auto-Seeding"; }
    QString description() const { return "Put a seed at local minima of mesh signal."; }
    QStringList parmNames() const { return QStringList() << QString("Radius (%1)").arg(UM); }
    QStringList parmDescs() const { return QStringList() 
      << "Size of neighborhood used for search of local minima. Typically, "
         "the radius of smallest cells in the sample.";
    }
    QStringList parmDefaults() const { return QStringList() << "3.0"; }
    QIcon icon() const { return QIcon(":/images/LocalMinima.png"); }
  };
  
  /**
   * \claass MeshNormalize MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * Local normalization of the mesh signal.
   */
  class mgxBase_EXPORT MeshNormalize : public MeshProcess {
  public:
    MeshNormalize(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_NON_EMPTY))
        return false;
      return (*this)(currentMesh(), parms[0].toFloat());
    }
  
    bool operator()(Mesh* mesh, float radius);
  
    QString folder() const { return "Signal"; }
    QString name() const { return "Normalize Signal"; }
    QString description() const { return "Normalize mesh signal locally."; }
    QStringList parmNames() const { return QStringList() << QString("Radius (%1)").arg(UM); }
    QStringList parmDescs() const { return QStringList() << "Size of neighborhood used for normalization."; }
    QStringList parmDefaults() const { return QStringList() << "5.0"; }
    QIcon icon() const { return QIcon(":/images/Normalize.png"); }
  };
  
  /**
   * \class MeshDilation MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * Morphological dilation on the mesh, using a spherical kernel. This process
   * replace the signal intensity on each vertex with the maximum within a
   * given distance.
   */
  class MeshDilation : public MeshProcess {
  public:
    MeshDilation(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_NON_EMPTY))
        return false;
      return (*this)(currentMesh(), parms[0].toFloat());
    }
  
    bool operator()(Mesh* mesh, float radius);
  
    QString folder() const { return "Signal"; }
    QString name() const { return "Dilate Signal"; }
    QString description() const { return "Morphological dilation of signal on mesh."; }
    QStringList parmNames() const { return QStringList() << QString("Radius (%1)").arg(UM); }
    QStringList parmDescs() const { return QStringList() << "Size of neighborhood used for dilation."; }
    QStringList parmDefaults() const { return QStringList() << "1.0"; }
    QIcon icon() const { return QIcon(":/images/Dilate.png"); }
  };
  
  /**
   * \class MeshErosion MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * Morphological erosion on the mesh, using a spherical kernel. This process
   * replace the signal intensity on each vertex with the minimum within a
   * given distance.
   */
  class MeshErosion : public MeshProcess {
  public:
    MeshErosion(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_NON_EMPTY))
        return false;
      return (*this)(currentMesh(), parms[0].toFloat());
    }
  
    bool operator()(Mesh* mesh, float radius);
  
    QString folder() const { return "Signal"; }
    QString name() const { return "Erode Signal"; }
    QString description() const { return "Apply morphological erosion to mesh signal (opposite to Dilate Signal)."; }
    QStringList parmNames() const { return QStringList() << QString("Radius (%1)").arg(UM); }
    QStringList parmDescs() const { return QStringList() << "Size of neighborhood used for erosion."; }
    QStringList parmDefaults() const { return QStringList() << "1.0"; }
    QIcon icon() const { return QIcon(":/images/Erode.png"); }
  };
  
  /**
   * \class MeshClosing MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * Morphological closing on the mesh. This process is equivalent to call the
   * MeshDilation followed by MeshErosion with the same size.
   */
  class MeshClosing : public MeshProcess {
  public:
    MeshClosing(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_NON_EMPTY))
        return false;
      return (*this)(currentMesh(), parms[0].toFloat());
    }
  
    bool operator()(Mesh* mesh, float radius);
  
    QString folder() const { return "Signal"; }
    QString name() const { return "Close Signal"; }
    QString description() const { return "Apply morphological dilation followed by erosion to mesh signal."; }
    QStringList parmNames() const { return QStringList() << QString::fromWCharArray(L"Radius (\xb5m)"); }
    QStringList parmDescs() const { return QStringList() << "Size of neighborhood used for dilation/erosion."; }
    QStringList parmDefaults() const { return QStringList() << "1.0"; }
    QIcon icon() const { return QIcon(":/images/Closing.png"); }
  };
  
  /**
   * \class MeshOpening MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * Morphological opening on the mesh. This process is equivalent to call the
   * MeshErosion followed by MeshDilation with the same size.
   */
  class MeshOpening : public MeshProcess {
  public:
    MeshOpening(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_NON_EMPTY))
        return false;
      return (*this)(currentMesh(), parms[0].toFloat());
    }
  
    bool operator()(Mesh* mesh, float radius);
  
    QString folder() const { return "Signal"; }
    QString name() const { return "Open Signal"; }
    QString description() const { return "Apply morphological erosion followed by dilation to mesh signal."; }
    QStringList parmNames() const { return QStringList() << QString::fromWCharArray(L"Radius (\xb5m)"); }
    QStringList parmDescs() const { return QStringList() << "Size of neighborhood used for erosion/dilation."; }
    QStringList parmDefaults() const { return QStringList() << "1.0"; }
    QIcon icon() const { return QIcon(":/images/Opening.png"); }
  };
  
  class mgxBase_EXPORT RescaleSignal : public MeshProcess {
  public:
    RescaleSignal(const MeshProcess& process) : Process(process) , MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(!checkState().mesh(MESH_NON_EMPTY))
        return false;
      bool ok;
      float percentile = parms[1].toFloat(&ok);
      if(!ok)
        return setErrorMessage("Error, argument 'Percentile' must be a number");
      float minimum = parms[2].toFloat(&ok);
      if(!ok)
        return setErrorMessage("Error, argument 'Minimum' must be a number");
      float maximum = parms[3].toFloat(&ok);
      if(!ok)
        return setErrorMessage("Error, argument 'Maximum' must be a number");
      return (*this)(currentMesh(), stringToBool(parms[0]), percentile, minimum, maximum);
    }
  
    bool operator()(Mesh* mesh, bool use_zero, float percentile, float minimum, float maximum);
  
    QString folder() const { return "Signal"; }
    QString name() const { return "Rescale Signal"; }
    QString description() const { return 
        "Change the colorbar of the signal.\n"
        "If percentile is set to 0, it uses the minimum and maximum arguments.\n"
        "Only the visualization is affected, the signal projection remains unchanged";
    }
    QStringList parmNames() const { return QStringList() 
      << "Zero as reference" << "Percentile" << "Minimum" << "Maximum";
    }
    QStringList parmDescs() const { return QStringList()
      << "If true, 0 will be used as a reference.\n"
         "If the signal is all positive (resp. negative), 0 will be added as a minimum (resp. maximum).\n"
         "If the signal is both positive and negative, 0 will be place at the center of the range"
      << "Keep only this percentage of the signal to compute the range."
      << "If the percentile specified is 0, uses this as the minimum value for the range"
      << "If the percentile specified is 0, uses this as the maximum value for the range";
    }
    QStringList parmDefaults() const { return QStringList() << "No" << "99" << "0" << "1"; }
  
    ParmChoiceMap parmChoice() const
    {
      ParmChoiceMap map;
      map[0] = booleanChoice();
      return map;
    }
  
    QIcon icon() const { return QIcon(":/images/Normalize.png"); }
  };
  
  /**
   * \class SetSignalProcess MeshProcessSignal.hpp <MeshProcessSignal.hpp>
   *
   * Set the signal to a given value for all currently active vertices.
   */
  class mgxBase_EXPORT SetSignalProcess : public MeshProcess {
  public:
    SetSignalProcess(const MeshProcess& process) : Process(process), MeshProcess(process) {}
  
    bool operator()(const QStringList& parms)
    {
      if(not checkState().mesh(MESH_VISIBLE))
        return false;
      bool ok;
      float value = parms[0].toFloat(&ok);
      if(not ok)
        return setErrorMessage("Parameter 'Value' must be a number");
      float percentile = parms[2].toFloat(&ok);
      if(not ok)
        return setErrorMessage("Parameter 'Percentile' must be a number");
      return (*this)(currentMesh(), value, stringToBool(parms[1]), percentile, stringToBool(parms[3]));
    }
  
    bool operator()(Mesh* m, float value, bool rescale, float percentile, bool use_zero);
  
    QString folder() const { return "Signal"; }
    QString name() const { return "Set Signal"; }
    QString description() const { return "Set the signal for the whole mesh, or for the currently selected part of it."; }
  
    QStringList parmNames() const { return QStringList() << "Value" << "Rescale" << "Percentile" << "Use zero"; }
    QStringList parmDescs() const { return QStringList() 
      << "New value for the signal"
      << "If true, the signal bounds will be rescaled"
      << "If rescaling, which percentile to use?"
      << "If rescaling, should we use zero as reference?";
    }
    QStringList parmDefaults() const { return QStringList() << "1" << "Yes" << "100" << "No"; }
    ParmChoiceMap parmChoice() const
    {
      ParmChoiceMap map;
      map[1] = booleanChoice();
      map[3] = booleanChoice();
      return map;
    }
    QIcon icon() const { return QIcon(":/images/Sharpen.png"); }
  };
  
  ///@}
  
  class mgxBase_EXPORT ProjectCurvatureOp {
  public:
    ProjectCurvatureOp(const vvgraph& S, const VtxVec& vs, std::string type, float radius, FloatVec& values,
                       std::vector<util::Curvature>& curvs);
    void operator()(int i);
  
  private:
    const vvgraph& S;
    const VtxVec& vs;
    std::string type;
    float radius;
    FloatVec& values;
    std::vector<util::Curvature>& curvs;
  };
  
  class mgxBase_EXPORT MeshGaussianBlurOp {
  public:
    MeshGaussianBlurOp(const vvgraph& S, const VtxVec& vs, float radius);
    void operator()(int i);
  
  private:
    const vvgraph& S;
    const VtxVec& vs;
    float radius;
  };
  
  class mgxBase_EXPORT MeshLocalMinimaOp {
  public:
    MeshLocalMinimaOp(const vvgraph& S, const VtxVec& vs, float radius, int startLabel);
    void operator()(int i);
  
  private:
    const vvgraph& S;
    const VtxVec& vs;
    float radius;
    int label;
  };
  
  class mgxBase_EXPORT MeshNormalizeOp {
  public:
    MeshNormalizeOp(const vvgraph& S, const VtxVec& vs, float radius);
  
    void operator()(int i);
  
  private:
    const vvgraph& S;
    const VtxVec& vs;
    float radius;
  };
  
  class MeshDilationOp {
  public:
    MeshDilationOp(const vvgraph& S, const VtxVec& vs, float radius);
    void operator()(int vi);
  
  private:
    const vvgraph& S;
    const VtxVec& vs;
    float radius;
  };
  
  class MeshErosionOp {
  public:
    MeshErosionOp(const vvgraph& S, const VtxVec& vs, float radius);
    void operator()(int vi);
  
  private:
    const vvgraph& S;
    const VtxVec& vs;
    float radius;
  };

}} 

#endif
