//
// This file is part of MorphoGraphX - http://www.MorphoGraphX.org
// Copyright (C) 2012-2015 Richard S. Smith and collaborators.
//
// If you use MorphoGraphX in your work, please cite:
//   http://dx.doi.org/10.7554/eLife.05864
//
// MorphoGraphX is free software, and is licensed under under the terms of the 
// GNU General (GPL) Public License version 2.0, http://www.gnu.org/licenses.
// 
#ifndef STACKPROCESSLABELS_HPP
#define STACKPROCESSLABELS_HPP

#include <Process.hpp>

namespace mgx {
namespace process {
///\addtogroup StackProcess
///@{
/**
 * \class WatershedStack StackProcessLabels.hpp <StackProcessLabels.hpp>
 *
 * Compute the seeded watershed of the current stack using the CImg library.
 */
class mgxBase_EXPORT WatershedStack : public StackProcess {
public:
  WatershedStack(const StackProcess& process)
    : Process(process)
    , StackProcess(process)
  {
  }

  bool operator()(const QStringList&)
  {
    if(!checkState().store(STORE_MAIN | STORE_VISIBLE | STORE_NON_LABEL).store(STORE_WORK | STORE_VISIBLE
                                                                               | STORE_LABEL))
      return false;
    Stack* stack = currentStack();
    Store* main = stack->main();
    Store* labels = stack->work();
    return (*this)(stack, main, labels);
  }

  bool operator()(Stack* stack, Store* main, Store* labels);

  QString folder() const {
    return "Segmentation";
  }
  QString name() const {
    return "Watershed3D";
  }
  QString description() const {
    return "3D Watershed on the current labeled stack.";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon(":/images/SegmentMesh.png");
  }
};

/**
 * \class ConsolidateRegions StackProcessLabels.hpp <StackProcessLabels.hpp>
 *
 * Try to merge segmented regions based on their contact area and the total voxel
 * intensity at their boundary.
 */
class mgxBase_EXPORT ConsolidateRegions : public StackProcess {
public:
  ConsolidateRegions(const StackProcess& process)
    : Process(process)
    , StackProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().store(STORE_MAIN | STORE_VISIBLE | STORE_NON_LABEL).store(STORE_WORK | STORE_VISIBLE
                                                                               | STORE_LABEL))
      return false;
    Stack* stack = currentStack();
    Store* main = stack->main();
    Store* labels = stack->work();
    return (*this)(main, labels, parms[0].toUInt(), parms[1].toUInt());
  }

  bool operator()(Store* data, Store* labels, uint threshold, uint minvoxels);

  QString folder() const {
    return "Segmentation";
  }
  QString name() const {
    return "Consolidate Regions";
  }
  QString description() const {
    return "Consilodate regions after watershed overseeding";
  }
  QStringList parmNames() const
  {
    return QStringList() << "Threshold"
                         << "Min Voxels";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "Threshold"
                         << "Min Voxels";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "5000"
                         << "100";
  }
  QIcon icon() const {
    return QIcon(":/images/Consolidate.png");
  }
};

/**
 * \class ConsolidateRegionsNormalized StackProcessLabels.hpp <StackProcessLabels.hpp>
 *
 * Variation on the ConsolidateRegions process.
 */
class mgxBase_EXPORT ConsolidateRegionsNormalized : public StackProcess {
public:
  ConsolidateRegionsNormalized(const StackProcess& process)
    : Process(process)
    , StackProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().store(STORE_MAIN | STORE_VISIBLE | STORE_NON_LABEL).store(STORE_WORK | STORE_VISIBLE
                                                                               | STORE_LABEL))
      return false;
    Stack* stack = currentStack();
    Store* main = stack->main();
    Store* labels = stack->work();
    return (*this)(main, labels, parms[0].toFloat());
  }

  bool operator()(Store* data, Store* labels, float tolerance);

  QString folder() const {
    return "Segmentation";
  }
  QString name() const {
    return "Consolidate Regions Normalized";
  }
  QString description() const {
    return "Consolidate regions with normalization (slower)";
  }
  QStringList parmNames() const {
    return QStringList() << "Tolerance";
  }
  QStringList parmDescs() const {
    return QStringList() << "Tolerance";
  }
  QStringList parmDefaults() const {
    return QStringList() << "0.3";
  }
  QIcon icon() const {
    return QIcon(":/images/Consolidate.png");
  }
};

/**
 * \class ThresholdLabelDelete StackProcessLabels.hpp <StackProcessLabels.hpp>
 *
 * Delete labels that have too few or too many voxels.
 */
class mgxBase_EXPORT ThresholdLabelDelete : public StackProcess {
public:
  ThresholdLabelDelete(const StackProcess& process)
    : Process(process)
    , StackProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().store(STORE_LABEL))
      return false;
    Stack* stack = currentStack();
    Store* input = stack->currentStore();
    Store* output = stack->work();
    if(!input)
      return false;
    bool res = (*this)(input, output, parms[0].toInt(), parms[1].toInt());
    if(res) {
      input->hide();
      output->show();
    }
    return res;
  }

  bool operator()(const Store* input, Store* labels, uint minVoxels, uint maxVoxels);

  QString folder() const {
    return "Segmentation";
  }
  QString name() const {
    return "Delete Labels by Threshold";
  }
  QString description() const {
    return "Delete Labels above/below voxel thresholds";
  }
  QStringList parmNames() const
  {
    return QStringList() << "Min voxels"
                         << "Max voxels";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "Min voxels"
                         << "Max voxels";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "1000"
                         << "0";
  }
  QIcon icon() const {
    return QIcon(":/images/DeleteLabel.png");
  }
};

/**
 * \class LocalMaximaStack StackProcessLabels.hpp <StackProcessLabels.hpp>
 *
 * Find the local maxima of the current image
 *
 * \note If the algorithm finds more than 65535 maxima, it will fail as it
 * cannot number them all.
 */
class mgxBase_EXPORT LocalMaximaStack : public StackProcess {
public:
  LocalMaximaStack(const StackProcess& process)
    : Process(process)
    , StackProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().store(STORE_NON_LABEL))
      return false;
    Stack* stack = currentStack();
    Store* input = stack->currentStore();
    Store* labels = stack->work();
    if(!input)
      return false;
    Point3f radius = Point3u(parms[0].toFloat(), parms[1].toFloat(), parms[2].toFloat());
    uint lab = parms[3].toUInt();
    uint minColor = parms[4].toUInt();
    bool res = (*this)(input, labels, radius, lab, minColor);
    if(res) {
      input->hide();
      labels->show();
    }
    return res;
  }

  bool operator()(const Store* input, Store* labels, Point3f radius, uint startLabel, uint minColor);

  QString folder() const {
    return "Segmentation";
  }
  QString name() const {
    return "Local Maxima";
  }
  QString description() const {
    return "Find local maxima and possibly number them";
  }
  QStringList parmNames() const
  {
    return QStringList() << QString("X Radius (%1)").arg(UM) << QString("Y Radius (%1)").arg(UM)
                         << QString("Z Radius (%1)").arg(UM) << "Start Label"
                         << "Min Color";
  }
  QStringList parmDescs() const
  {
    return QStringList() << QString("X Radius (%1)").arg(UM) << QString("Y Radius (%1)").arg(UM)
                         << QString("Z Radius (%1)").arg(UM) << "Start Label"
                         << "Min Color";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "5.0"
                         << "5.0"
                         << "5.0"
                         << "2"
                         << "10000";
  }
  QIcon icon() const {
    return QIcon(":/images/LocalMaxima.png");
  }
};

/**
 * \class FillLabelStack StackProcessLabels.hpp <StackProcessLabels.hpp>
 *
 * Replace one label by another one on the stack
 */
class mgxBase_EXPORT FillLabelStack : public StackProcess {
public:
  FillLabelStack(const StackProcess& process)
    : Process(process)
    , StackProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().store(STORE_LABEL))
      return false;
    Stack* stack = currentStack();
    Store* input = stack->currentStore();
    Store* output = stack->work();
    if(!input)
      return false;
    bool res = (*this)(input, output, parms[0].toUShort(), parms[1].toUShort());
    if(res) {
      input->hide();
      output->show();
    }
    return res;
  }

  bool operator()(const Store* input, Store* output, ushort filledLabel, ushort newLabel);

  QString folder() const {
    return "Segmentation";
  }
  QString name() const {
    return "Fill Label";
  }
  QString description() const {
    return "Replace a label with another one";
  }
  QStringList parmNames() const
  {
    return QStringList() << "Filled label"
                         << "New label";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "Filled label"
                         << "New label";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "1000"
                         << "0";
  }
  QIcon icon() const {
    return QIcon(":/images/FillLabel.png");
  }
};

/**
 * \class EraseAtBorderStack StackProcessLabels.hpp <StackProcessLabels.hpp>
 *
 * Erase any label touching the border of the image.
 */
class mgxBase_EXPORT EraseAtBorderStack : public StackProcess {
public:
  EraseAtBorderStack(const StackProcess& process)
    : Process(process)
    , StackProcess(process)
  {
  }

  bool operator()(const QStringList&)
  {
    if(!checkState().store(STORE_LABEL))
      return false;
    Stack* stack = currentStack();
    Store* input = stack->currentStore();
    Store* output = stack->work();
    if(!input)
      return false;
    bool res = (*this)(input, output);
    if(res) {
      input->hide();
      output->show();
    }
    return res;
  }

  bool operator()(const Store* input, Store* output);

  QString folder() const {
    return "Segmentation";
  }
  QString name() const {
    return "Erase at Border";
  }
  QString description() const {
    return "Erase any labelled region touching the border of the image";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QStringList parmDefaults() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon(":/images/EraseLabel.png");
  }
};
///@}
} // namespace process
} // namespace mgx

#endif
